from flask import Blueprint, request, redirect, url_for
from flask_restful import Api
from flask import current_app

bp = Blueprint('auth', __name__, url_prefix='/auth')
api = Api(bp)


def login_response(uid, sid):
    from flask import make_response
    next = request.args.get('next')
    response = make_response(redirect(next) if next else redirect(url_for('index')))
    response.set_cookie('uid', uid)
    response.set_cookie('sid', sid)
    return response


@bp.route('/login', methods=('POST',))
def login():
    username = request.args.get('username', None)
    password = request.args.get('password', None)

    uid = current_app.auth_manager.check_auth(username, password)
    if not uid:
        return current_app.auth_manager.authenticate()

    sid = current_app.auth_manager.login_user(uid, username=username, password=password)

    from flask import make_response
    response = make_response("login succeeded!", 200)
    response.set_cookie('uid', uid)
    response.set_cookie('sid', sid)
    return response


@bp.route('/login-view', methods=('GET',))
def login_view():
    auth = request.authorization
    if not auth:
        return current_app.auth_manager.authenticate()
    uid = current_app.auth_manager.check_auth(auth.username, auth.password)
    if not uid:
        return current_app.auth_manager.authenticate()

    sid = current_app.auth_manager.login_user(uid, username=auth.username, password=auth.password)

    return login_response(uid, sid)


@bp.route('/login-redirect', methods=('POST',))
def login_redirect():
    username = request.form.get('username', None)
    password = request.form.get('password', None)

    uid = current_app.auth_manager.check_auth(username, password)
    if not uid:
        return current_app.auth_manager.authenticate()

    sid = current_app.auth_manager.login_user(uid, username=username, password=password)

    return login_response(uid, sid)



from functools import wraps

from flask import Response, request, current_app
from flask.sessions import SecureCookieSessionInterface
from flask_login import UserMixin


class ParadeUser(UserMixin):
    pass


class DisabledSessionInterface(SecureCookieSessionInterface):
    """Prevent creating session from API requests."""

    def save_session(self, *args, **kwargs):
        return


def auth_required(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        auth_manager = current_app.auth_manager
        auth_token = request.args.get('sid') or request.cookies.get('sid')
        if not auth_token or not auth_manager.check_token(auth_token):
            return auth_manager.authenticate()
        return f(*args, **kwargs)

    return decorated


class AuthManager(object):
    _login_user = dict()

    def check_auth(self, user_login_key, secret):
        """
        This function is called to check if a user/secret combination is valid.
        :param user_login_key: the login_key to login user
        :param secret: the secret to login user
        :return: the user_key of login-user or None if check failed
        """
        import hashlib
        hl = hashlib.md5()
        hl.update('parade'.encode(encoding='utf-8'))
        if user_login_key == 'parade' and secret == hl.hexdigest():
            return 'parade'

        return None

    def login_user(self, user_key, **kwargs):
        """
        This function put the login-user into session and return the auth token
        :param user_key: the key of login-user
        :param kwargs: the extra arguments about the login-user
        :return: the auth token put in session
        """
        import uuid
        token = str(uuid.uuid3(uuid.NAMESPACE_OID, user_key))
        self._login_user[user_key] = token
        return token

    def check_token(self, user_key, auth_token):
        """
        This function check the auth token per request
        :param user_key: the user key the request announced issue-from
        :param auth_token: the auth token of the login-user
        :return: the loaded user if success otherwise None
        """
        if user_key and auth_token and user_key in self._login_user and self._login_user[user_key] == auth_token:
            user = ParadeUser()
            user.id = user_key
            user.token = auth_token
            return user
        return None

    def authenticate(self):
        """Sends a 401 response that enables basic auth"""
        return Response(
            'Could not verify your access level for that URL.\n'
            'You have to login with proper credentials', 401)



# -*- coding:utf-8 -*-
from ..server import start_webapp
from . import ParadeCommand


class ServerCommand(ParadeCommand):
    requires_workspace = True

    def run_internal(self, context, **kwargs):
        port = int(kwargs.get('port', 5000))
        enable_dash = kwargs.get('enable_dash')
        enable_static = kwargs.get('enable_static')
        enable_socketio = kwargs.get('enable_socketio')
        start_webapp(context, port=port, enable_static=enable_static, enable_dash=enable_dash, enable_socketio=enable_socketio)

    def short_desc(self):
        return 'start a parade api server'

    def config_parser(self, parser):
        parser.add_argument('-p', '--port', default=5000, help='the port of parade server')
        parser.add_argument('--enable-static', action="store_true", help='enable static template rendering'
                                                                         'support in parade server')
        parser.add_argument('--enable-dash', action="store_true", help='enable dash support in parade server')
        parser.add_argument('--enable-socketio', action="store_true", help='enable socketio support in parade server')



# -*- coding:utf-8 -*-

import dash_html_components as html
import dash
from ..core.context import Context


class Dashboard(object):

    def __init__(self, app: dash.Dash, context: Context):
        self.app = app
        self.context = context

    @property
    def name(self):
        """
        get the identifier of the task, the default is the class name of task
        :return: the task identifier
        """
        return self.__module__.split('.')[-1]

    @property
    def display_name(self):
        return self.name

    @property
    def layout(self):
        return html.Div([html.H1('Content of dashboard [' + self.name + ']')])



# -*- coding:utf-8 -*-
from ..server.auth import DisabledSessionInterface, AuthManager
from .dashboard import Dashboard
from ..core.context import Context
from ..utils.modutils import iter_classes, walk_modules


def load_dashboards(app, context, name=None):
    """
    generate the task dict [task_key => task_obj]
    :return:
    """
    d = {}
    for dash_class in iter_classes(Dashboard, context.name + '.dashboard'):
        dashboard = dash_class(app, context)
        dash_name = dashboard.name
        if name and dash_name != name:
            continue
        d[dash_name] = dashboard
    return d


def load_contrib_apis(app, context):
    from importlib import import_module
    try:
        import_module(context.name + '.api')
    except:
        return

    for api_module in walk_modules(context.name + '.api'):
        try:
            app.register_blueprint(api_module.bp)
        except:
            pass


def _load_dash(app, context):
    import dash_html_components as html
    import dash_core_components as dcc
    from dash.dependencies import Input, Output

    dashboards = load_dashboards(app, context)
    dashboard_opts = [{'label': dashboards[dashkey].display_name, 'value': dashkey} for dashkey in dashboards]
    default_dashboard = None
    # if len(dashboard_opts) > 0:
    #     default_dashboard = dashboard_opts[0]['value']

    app.layout = html.Div(
        [
            dcc.Location(id='url', refresh=False),
            # header
            html.Div([

                html.Span("Parade Dashboard", className='app-title four columns', style={"marginTop": "8px"}),

                html.Div([
                    dcc.Dropdown(
                        id="tabs",
                        options=dashboard_opts,
                        value=default_dashboard
                    )], className="two columns", style={"marginTop": "16px"}),

                # # title content
                # html.Div(id="title_placeholder", className="two columns", style={"marginTop": "16px"}),

                html.Div(
                    html.Img(
                        src='https://s3-us-west-1.amazonaws.com/plotly-tutorials/logo/new-branding/dash-logo-by-plotly-stripe-inverted.png',
                        height="100%")
                    , style={"float": "right", "height": "100%"})
            ],
                className="row header"
            ),

            # Tab content
            html.Div(id="tab_content", className="row", style={"margin": "2% 3%"}),

            html.Link(
                href="https://cdn.rawgit.com/plotly/dash-app-stylesheets/2d266c578d2a6e8850ebce48fdb52759b2aef506/stylesheet-oil-and-gas.css",
                rel="stylesheet"),
            html.Link(
                href="https://cdn.rawgit.com/amadoukane96/8a8cfdac5d2cecad866952c52a70a50e/raw/cd5a9bf0b30856f4fc7e3812162c74bfc0ebe011/dash_crm.css",
                rel="stylesheet"),
        ],
        className="row",
        style={"margin": "0%"},
    )

    @app.callback(Output("tab_content", "children"),
                  [Input('url', 'pathname'),
                   Input('tabs', 'value')])
    def render_content(path, dropdown_tab):
        if not path:
            return html.Div([html.H1("Please select the dashboard")])
        path_tab = path[len('/dashboard/'):]
        if len(path_tab) == 0:
            path_tab = None

        tab = dropdown_tab or path_tab
        if tab in dashboards:
            return dashboards[tab].layout
        else:
            return html.Div([html.H1("Please select the dashboard")])


def _init_web(context, enable_auth):
    from flask import Blueprint
    from flask import render_template
    from flask_login import login_required
    import os

    web = Blueprint('web', __name__)

    template_dir = os.path.join(context.workdir, 'web')
    index_file = os.path.join(template_dir, 'index.html')
    login_file = os.path.join(template_dir, 'login.html')

    if os.path.exists(index_file):
        if enable_auth:
            @web.route("/")
            @login_required
            def route_auth():
                return render_template("index.html")
        else:
            @web.route("/")
            def route():
                return render_template("index.html")

    if os.path.exists(login_file) and enable_auth:
        @web.route("/login")
        def login():
            return render_template("login.html")

    return web


def _init_socketio(app, context):
    from flask_socketio import SocketIO
    socketio = SocketIO(app, async_mode='threading')
    sio = socketio.server

    @sio.on('connect', namespace='/exec')
    def connect(sid, environ):
        pass

    @sio.on('query', namespace='/exec')
    def query(sid, data):
        exec_id = data
        sio.enter_room(sid, str(exec_id), namespace='/exec')
        sio.emit('reply', exec_id, namespace='/exec')

    socketio.init_app(app)


def _init_auth(app, context):
    app.secret_key = 'parade'
    from flask_login import LoginManager
    login_manager = LoginManager()
    login_manager.login_view = "/auth/login-view"
    login_manager.login_view = "/login"
    login_manager.init_app(app)

    if not context.conf.has('auth.driver'):
        app.auth_manager = AuthManager()
    else:
        from ..utils.modutils import get_class
        auth_cls = get_class(context.conf['auth.driver'], AuthManager, context.name + '.contrib.auth')
        app.auth_manager = auth_cls()

    @login_manager.request_loader
    def load_user_by_request(request):
        user_key = request.args.get('uid') or request.cookies.get('uid')
        auth_token = request.args.get('sid') or request.cookies.get('sid')
        return app.auth_manager.check_token(user_key, auth_token)

    app.session_interface = DisabledSessionInterface()
    from .auth import api as auth_api
    app.register_blueprint(auth_api.bp)


def start_webapp(context: Context, port=5000, enable_auth=True, enable_static=False, enable_dash=False,
                 enable_socketio=True):
    import os
    from flask import Flask
    from flask_cors import CORS

    template_dir = os.path.join(context.workdir, 'web')
    static_dir = os.path.join(context.workdir, 'web', 'static')

    app = Flask(context.name, template_folder=template_dir, static_folder=static_dir)
    CORS(app)

    app.parade_context = context

    from parade.server.api import parade_blueprint
    app.register_blueprint(parade_blueprint)

    load_contrib_apis(app, context)

    if enable_auth:
        _init_auth(app, context)

    if enable_static or enable_dash:
        web_blueprint = _init_web(context, enable_auth)
        app.register_blueprint(web_blueprint)

    if enable_socketio:
        _init_socketio(app, context)
    context.webapp = app
    debug = context.conf.get_or_else('debug', False)

    if enable_dash:
        import dash
        app_dash = dash.Dash(__name__, server=app, url_base_pathname='/dashboard/')
        app_dash.config.suppress_callback_exceptions = True

        _load_dash(app_dash, context)

        def protect_views(app):
            from flask_login import login_required
            for view_func in app.server.view_functions:
                if view_func.startswith(app.url_base_pathname):
                    app.server.view_functions[view_func] = login_required(app.server.view_functions[view_func])

            return app

        if enable_auth:
            app_dash = protect_views(app_dash)

        @app.route("/dashboard")
        def route_dash():
            import flask
            return flask.redirect('/dash')

        from werkzeug.wsgi import DispatcherMiddleware
        app = DispatcherMiddleware(app, {
            '/dash': app_dash.server
        })

        from werkzeug.serving import run_simple
        run_simple('0.0.0.0', port, app, use_reloader=True, use_debugger=debug)
    else:
        app.run(host="0.0.0.0", port=port, debug=debug)



#!/usr/bin/env python
# -*- coding:utf-8 -*-

from setuptools import setup, find_packages

classifiers = """\
Development Status :: 4 - Beta
Intended Audience :: Developers
License :: OSI Approved :: Apache Software License
Programming Language :: Python :: 3
Topic :: Database
Topic :: Software Development :: Libraries :: Python Modules
Operating System :: Unix
Operating System :: MacOS :: MacOS X
Operating System :: Microsoft :: Windows
Operating System :: POSIX
"""

setup(
    name='parade',
    # auto generate version
    # use_scm_version=True,
    version='0.1.22.5',
    author='He Bai',
    author_email='bailaohe@gmail.com',

    description='A ETL engine for dataframe based data task',

    keywords=["parade", "ETL", "dataframe", "schedule", "database"],
    url='https://github.com/bailaohe/parade',
    platforms=["any"],
    classifiers=filter(None, classifiers.split("\n")),

    install_requires=['pandas', 'sqlalchemy', 'requests', 'pyyaml',
                      'psycopg2', 'mysqlclient', 'xlsxwriter',
                      'flask', 'flask_cors', 'flask_restful', 'Flask-SocketIO',
                      'PyGithub', 'pymongo', 'redis', 'jupyter', 'nbformat'],

    packages=find_packages('src'),
    package_dir=({'parade': 'src/parade'}),
    zip_safe=False,

    include_package_data=True,
    package_data={'': ['*.json', '*.xml', '*.yml', '*.tpl']},

    entry_points={
        'console_scripts': [
            'parade = parade.cmdline:execute',
        ],
    },

    extras_require={
        "feature": ["parade-feature"],
    },
    # setup_requires=[
    #     "setuptools_scm>=1.5",
    # ],
    python_requires=">=3.4",
    # download_url='https://github.com/bailaohe/parade/tarball/0.1',
)
